#  Copyright (c) 2020 SBA - MIT License

from unittest import TestCase
from unittest.mock import Mock

import io
import time

from non_blocking_io_wrapper.non_blocking_reader import NonBlockingReader


class TimedRead:
    def __init__(self, *seq):
        self.seq = seq[:]
        self.it = iter(self.seq)

    def __call__(self, _hint=-1):
        try:
            delay, data = next(self.it)
        except StopIteration:
            return b''
        time.sleep(delay)
        return data


class TestNonBlockingReader(TestCase):
    def setUp(self) -> None:
        raw = Mock(io.RawIOBase)
        raw.read = Mock()
        raw.read.side_effect = TimedRead((.1, b'a'), (.1, b'b\nc'), (.1, b'd'))
        self.fd = NonBlockingReader(raw)

    def test_drain(self):
        self.fd.drain()
        self.assertEqual(b'ab\ncd', self.fd.read())

    def test_timed_read(self):
        self.assertIsNone(self.fd.read())
        time.sleep(.25)
        self.assertEqual(b'ab\nc', self.fd.read())

    def test_readline_no_data(self):
        with self.assertRaises(OSError):
            self.fd.readline()

    def test_readline_no_eol(self):
        time.sleep(.15)
        data = self.fd.readline()
        self.assertEqual(b'a', data)

    def test_readline_eol(self):
        time.sleep(.25)
        data = self.fd.readline()
        self.assertEqual(b'ab\n', data)

    def test_iter(self):
        self.fd.drain()
        self.assertEqual([b'ab\n', b'cd'], list(self.fd))

    def test_select_timeout(self):
        self.assertFalse(self.fd.select(.05))
        self.assertIsNone(self.fd.read())

    def test_select_ok(self):
        beg = time.time()
        self.assertTrue(self.fd.select(.5))
        end = time.time()
        self.assertEqual(b'a', self.fd.read())
        self.assertTrue(.05 < end-beg < .15)

    def test_select_eof(self):
        self.fd.drain()
        self.assertEqual(b'ab\ncd', self.fd.read())
        beg = time.time()
        self.assertTrue(self.fd.select(.1))
        end = time.time()
        self.assertTrue(end-beg < .05)
        self.assertEqual(b'', self.fd.read())



#  Copyright (c) 2020 SBA - MIT License

import io
import threading
from typing import Optional


class NonBlockingReader(io.RawIOBase):
    """Lightweight wrapper to create a non blocking version of any (byte) stream.

    It is a true io.RawIOBase subclass. Because of that, any method valid
    on an input byte stream should work seamlessly. Yet, a line access
    (readline, or iterator) will raise an OSError if no data is available at
    call time (you can use the `select` method to test/wait for
    that condition). If some data is available a line will be returned, or a
    partial one if no newline is available in the buffer.

    If no data is present, a read will return None instead of blocking.
    """
    def __init__(self, stream):
        self.stream = stream
        self.t = threading.Thread(target=self._get_data, name='get_data',
                                  daemon=True)
        self.lock = threading.Lock()
        self.available = threading.Event()
        self.data = []
        self.ended = False
        self.t.start()

    def readable(self) -> bool:
        return True

    def readinto(self, __buffer: bytearray) -> Optional[int]:
        with self.lock:
            if len(self.data) == 0:
                return 0 if self.ended else None
            begin = 0
            mx = len(__buffer)
            for i, block in enumerate(self.data):
                if mx == 0:
                    break
                if len(block) < mx:
                    ln = len(block)
                    __buffer[begin: begin + ln] = block
                    begin += ln
                    mx -= ln
                else:
                    __buffer[begin:] = block[:mx]
                    self.data[i] = block[mx:]
                    begin += mx
                    i -= 1
                    break
            if i >= 0:
                del self.data[:i + 1]
            if len(self.data) == 0:
                self.available.clear()
            return begin

    def _get_data(self):
        """A thread that continuously reads the underlying stream."""
        while not (self.closed or self.ended):
            block = self.stream.read(8192)
            with self.lock:
                if len(block) == 0:
                    self.ended = True
                else:
                    self.data.append(block)
                    self.available.set()

    def drain(self, timeout=None):
        """Waits for the underlying stream to reach its end of file.

        If timeout is given and is not None it is the maximum time in
        seconds to wait.
        Returns True if it succeeded or False on timeout.
        """
        self.t.join(timeout)
        return not self.t.is_alive()

    def __del__(self):
        self.close()
        super().__del__()

    def select(self, timeout=None):
        """Waits for data to become available, or more exactly for the
        next read not to return None. Specifically, it returns True after
        end of file has been reached, because the next read will
        immediately return b''.

        If timeout is given and is not None it is the maximum time in
        seconds to wait.
        Returns True if it succeeded or False on timeout.
        """
        if self.ended:
            return True
        self.available.wait(timeout)
        with self.lock:
            return 0 != len(self.data)



#  Copyright (c) 2020 SBA - MIT License

from setuptools import setup
from warnings import warn


import os.path
import re
import subprocess

import sys

sys.path.append('.')

pkg = 'non_blocking_io_wrapper'
wd = os.path.abspath(os.path.dirname(__file__))
_version = None


def get_version() -> str:
    global _version
    if _version is not None:
        return _version
    """ extract version number """
    _version = '0.0.0'  # fallback value should never be used
    try:  # first from git using setuptools_scm
        from setuptools_scm import get_version as scm_version
        print('found setuptools_scm')
        _version = scm_version(write_to=os.path.join(wd, pkg, 'version.py'))
    except (ImportError, LookupError):
        try:  # else from a previous version.py
            with open(os.path.join(wd, pkg, 'version.py')) as fd:
                for line in fd:
                    if line.startswith('version'):
                        _version = line.split("'")[1]
        except OSError:
            warn('Need either git+setuptools-scm or version.py file')
    return _version


def get_commit() -> str:
    try:
        p = subprocess.run('git show --format=%H -s', capture_output=True,
                           check=True, shell=True, encoding='Latin1')
    except (OSError, subprocess.CalledProcessError):
        return ''
    return p.stdout.strip()


def get_long_desc() -> str:
    """ read long description and adjust master with version for badges or links
    only for release versions (x.y.z)
    """
    get_version()
    release = re.compile(r'(\d+\.){0,2}\d+$')
    with open(os.path.join(wd, 'README.md')) as fd:
        if _version == '0.0.0' or not release.match(_version):
            _long_description = fd.read()
        else:
            lines = fd.readlines()
            for i, line in enumerate(lines):
                if not line.startswith('['):
                    break
                if 'travis' in line:
                    lines[i] = line.replace('master', _version)
                elif 'codecov' in line:
                    commit = get_commit()
                    if commit != '':
                        lines[i] = line.replace('branch/master',
                                                'commit/' + commit)
            _long_description = ''.join(lines)
    return _long_description


if __name__ == '__main__':
    setup(
        version=get_version(),
        long_description=get_long_desc(),
        long_description_content_type='text/markdown',
    )
